import sys
from .sugar import TriggerMode, Reader, LogLevel  # noqa: F401

# Use either the real or mock (placeholder) implementations of the
# Redis Gears Python environment top level builtin funvtions
# Depending on if the module is loaded in a 'redisgears' environment
# or not (e.g. dev or client)
if "redisgears" in sys.modules:
    # Server Gears runtime environment
    # Import the default functions and classes
    from redisgears import atomicCtx as atomic  # noqa: F401
    from redisgears import executeCommand as execute  # noqa: F401
    from redisgears import getMyHashTag as hashtag  # noqa: F401
    from redisgears import log  # noqa: F401
    from __main__ import gearsConfigGet  # noqa: F401
    from __main__ import GearsBuilder  # noqa: F401
    from __main__ import GB  # noqa: F401
else:
    # Dev or Client environment
    # Import placeholder functions and classes
    from .placeholders import (  # noqa: F401
        atomic,
        execute,
        hashtag,
        log,
        configGet,
        gearsConfigGet,
        GearsBuilder,
        GB,
    )



import logging
import redgrease.operations as gearop
from redgrease.sugar import Reader, TriggerMode, LogLevel
from redgrease.typing import (
    T,
    Key,
    Callback,
    Processor,
    Filterer,
    Accumulator,
    Reducer,
    BatchReducer,
    Extractor,
    Mapper,
    Expander,
)


logger = logging.getLogger(__name__)


class atomic:
    def __enter__(self):
        pass

    def __exit__(self, ex_type, ex_value, ex_traceback):
        pass


def execute(command: str, *args) -> str:
    """execute redis command"""
    # TODO: Call some default Redis Server?
    pass


def hashtag() -> str:
    pass


def log(message: str, level: str = LogLevel.Notice):
    logger.log(level=LogLevel.to_logging_level(level), msg=message)


def configGet(key: str):
    pass


def gearsConfigGet(key: str, default=None):
    pass


class GearsBuilder:
    def __init__(
        self,
        reader: str = Reader.KeysReader,
        defaultArgs: str = "*",
        desc: str = None,
    ):
        """Gear process

        Args:
            reader (str, optional): Input records reader
            Defining where the input to the gear will come from.
            One of:
            - 'KeysReader':
            - 'KeysOnlyReader':
            - 'StreamReader':
            - 'PythonReader':
            - 'ShardsReader':
            - 'CommandReader':
            Defaults to 'KeysReader'.

            defaultArgs (str, optional): Additional arguments
            to the reader.
            These are usually a key's name, prefix, glob-like
            or a regular expression.
            Its use depends on the function's reader type and action.
            Defaults to '*'.

            desc (str, optional): An optional description.
            Defaults to None.
        """
        self.reader = reader
        self.defaultArgs = defaultArgs
        self.desc = desc
        self.operations = []

    def run(
        self, arg: str = None, convertToStr: bool = True, collect: bool = True, **kargs
    ):
        """Runs a gear function as a batch.
        The function is executed once and exits once the data is
        exhausted by its reader.

        Args:
            arg (str, optional): An optional argument that's passed to
            the reader as its defaultArg.
            It means the following:
            - A glob-like pattern for the KeysReader
            and KeysOnlyReader readers.
            - A key name for the StreamReader reader.
            - A Python generator for the PythonReader reader.

            Defaults to None.

            convertToStr (bool, optional): When `True` adds a map
            operation to the flow's end that stringifies records.
            Defaults to True.

            collect (bool, optional): When `True` adds a collect operation
            to flow's end.
            Defaults to True.

        Returns:
            [type]: [description]
        """
        self.operations.append(
            gearop.Run(arg=arg, convertToStr=convertToStr, collect=collect, kargs=kargs)
        )
        return self

    def register(
        self,
        prefix: str = "*",
        convertToStr: bool = True,
        collect: bool = True,
        mode: str = TriggerMode.Async,
        onRegistered: Callback = None,
        trigger: str = None,
        **kargs
    ):
        """Runs a Gear function as an event handler.
        The function is executed each time an event arrives.
        Each time it is executed, the function operates on the event's
        data and once done is suspended until its future invocations by
        new events.

        Args:
            prefix (str, optional): Key prefix pattern to match on.
            Not relevant for 'CommandReader' readers (see 'trigger').
            Defaults to '*'.

            convertToStr (bool, optional): When `True` adds a map
            operation to the flow's end that stringifies records.
            Defaults to True.

            collect (bool, optional): When True adds a collect operation
            to flow's end.
            Defaults to True.

            mode (str, optional): The execution mode of the function.
            Can be one of:
            - 'async': Execution will be asynchronous across the entire
            cluster.
            - 'async_local': Execution will be asynchronous and restricted
            to the handling shard.
            - 'sync': Execution will be synchronous and local.

            Defaults to 'async'.

            onRegistered (Callback, optional): A function callback that's
            called on each shard upon function registration.
            It is a good place to initialize non-serializable objects
            such as network connections.
            Defaults to None.

            trigger (str, optional): For 'CommandReader' only.
            The trigger string that will trigger the function.
            Defaults to None.

        Returns:
            [type]: [description]
        """
        self.operations.append(
            gearop.Register(
                prefix=prefix,
                convertToStr=convertToStr,
                collect=collect,
                mode=mode,
                onRegistered=onRegistered,
                trigger=trigger,
                kargs=kargs,
            )
        )
        return self

    def map(self, op: Mapper):
        """Instance-local Map operation that performs a one-to-one
        (1:1) mapping of records.

        Args:
            op (Mapper): Mapper function (1:1)

        Returns:
            A gear flow outputting the transformed records
        """
        self.operations.append(gearop.Map(op=op))
        return self

    def flatmap(self, op: Expander):
        """Instance-local FlatMap operation that performs one-to-many
        (1:N) mapping of records.

        It requires one expander callback that maps a single input record
        to one or more output records.

        FlatMap is nearly identical to the Map operation.
        Unlike regular mapping, however, when FlatMap returns a list,
        each element in the list is turned into a separate output record.

        Args:
            op (Expander): Expader function (1:N)

        Returns:
            A gear flow outputting the expanded records
        """
        self.operations.append(gearop.FlatMap(op=op))
        return self

    def foreach(self, op: Processor):
        """Instance-local ForEach operation performs one-to-the-same
        (1=1) mapping.

        It requires one processor callback to perform some work that's
        related to the input record.

        Its output record is a copy of the input, which means anything
        the callback returns is discarded

        Args:
            op (Processor): Processor functon to be used on each record.

        Returns:
            A gear flow outputting the **input** records unmodified
        """
        self.operations.append(gearop.ForEach(op=op))
        return self

    def filter(self, op: Filterer):
        """Instance-local Filter operation performs one-to-zero-or-one
        (1:bool) filtering of records.

        It requires a filterer function callback.

        An input record that yields a falsehood will be discarded and
        only truthful ones will be output.

        Args:
            op (Filterer): Filteriong / predicate function

        Returns:
            A gear flow outputting only the records tha pass the
            filter (i.e evaluates to 'True')
        """
        self.operations.append(gearop.Filter(op=op))
        return self

    def accumulate(self, op: Accumulator):
        """Instance-local Accumulate operation performs many-to-one
        mapping (N:1) of records.

        It requires one accumulator callback.

        Once input records are exhausted its output is a single record
        consisting of the accumulator's value.

        Args:
            op (Accumulator): Accumulator function

        Returns:
            A gear flow outputting the final accumulated value
        """
        self.operations.append(gearop.Accumulate(op=op))
        return self

    def localgroupby(self, extractor: Extractor[Key], reducer: Reducer):
        """Instance-local LocalGroupBy operation performs many-to-less
        mapping (N:M) of records.

        The operation requires two callbacks: an extractor and a reducer.

        The output records consist of the grouping key and its respective
        accumulator's value.

        Args:
            extractor (Extractor[Key]): Extractor funtction that extracts
            the key that will be used to group by, from each record.

            reducer (Reducer): Reducer function that reduce the records of
            each 'group' into a single value for the group.

        Returns:
            A gear flow outputting each group and its reduced value
        """
        self.operations.append(
            gearop.LocalGroupBy(extractor=extractor, reducer=reducer)
        )
        return self

    def limit(self, length: int, start: int = 0):
        """Instance-local Limit operation limits the number of records.

        It accepts two numeric arguments: a starting position in the input
        records "array" and a maximal number of output records.

        Args:
            length (int): Maximum number of outputted records.

            start (int, optional): 0-based offset / index of the first
            record that output should start from.
            Defaults to 0.

        Returns:
            A gear flow outputting the records between the start
            record index (inclusive) and start+length record index
            (exclusive).
        """
        self.operations.append(gearop.Limit(length=length, start=start))
        return self

    def collect(self):
        """Cluster-global Collect operation collects the result records
        from all of the shards to the originating one.

        It has no arguments.

        Returns:
            A gear flow outputting the input records across all shards.
        """
        self.operations.append(gearop.Collect())
        return self

    def repartition(self, extractor: Extractor[Key]):
        """Cluster-global Repartition operation repartitions the records
        by them shuffling between shards.

        It accepts a single key extractor function callback.
        The extracted key is used for computing the record's new placement
        in the cluster (i.e. hash slot).
        The operation then moves the record from its original shard
        to the new one.

        Args:
            extractor (Extractor[Key]): [description]

        Returns:
            A gear flow outputting the input records, but each record
            is repartitioned to the shards as per the extractor function.
        """
        self.operations.append(gearop.Repartition(extractor=extractor))
        return self

    def aggregate(self, zero: T, seqOp: Accumulator[T], combOp: Accumulator[T]):
        """Perform aggregation on all the execution data.

        Args:
            zero (T): The first value that will pass to the aggregation
            function.

            seqOp (Accumulator[T]): The local aggregate function
            (will be performed on each shard)

            combOp (Accumulator[T]): The global aggregate function
            (will be performed on the results of seqOp from each shard)

        Returns:
            A gear flow with the output of the
            'combOp' aggregation as output.
        """
        self.operations.append(gearop.Aggregate(zero=zero, seqOp=seqOp, combOp=combOp))
        return self

    def aggregateby(
        self, extractor: Extractor[Key], zero: T, seqOp: Reducer[T], combOp: Reducer[T]
    ):
        """Like aggregate, but on each key, the key is extracted using the extractor.

        Args:
            extractor (Extractor[Key]): A function that get as input the
            record and return the aggregated key.

            zero (T): The first value that will pass to the aggregation
            function.

            seqOp (Reducer[T]): The local aggregate function
            (will be performed on each shard)

            combOp (Reducer[T]): The global aggregate function
            (will be performed on the results of seqOp from each shard)

        Returns:
            [T]: The gear flow with the output of the
            'combOp' aggregation as output.
        """
        self.operations.append(
            gearop.AggregateBy(
                extractor=extractor, zero=zero, seqOp=seqOp, combOp=combOp
            )
        )
        return self

    def groupby(self, extractor: Extractor[Key], reducer: Reducer[T]):
        """performs a many-to-less (N:M) grouping of records.
        It is similar to AggregateBy but uses only a global reducer.
        It can be used in cases where locally reducing the data isn't
        possible.

        The operation requires two callbacks: an extractor a reducer .

        The operation is made of these steps:
        1. A global repartition operation that uses the extractor
        2. The reducer is locally invoked

        Args:
            extractor (Extractor[Key]): Key extraction function

            reducer (Reducer[T]): Reducer function

        Returns:
            A gear flow outputting a locally-reduced list of records,
            one for each key.
            The output records consist of the grouping key and its
            respective accumulator's value.
        """
        self.operations.append(gearop.GroupBy(extractor=extractor, reducer=reducer))
        return self

    def batchgroupby(self, extractor: Extractor[Key], reducer: BatchReducer[T]):
        """Many-to-less (N:M) grouping of records.

        Note: Using this operation may cause a substantial increase in
        memory usage during runtime.
        Instead of using BatchGroupBy, prefer using the GroupBy
        operation as it is more efficient and performant.
        Only use BatchGroupBy when the reducer's logic requires the full
        list of records for each input key.

        The operation requires two callbacks: an extractor a batch reducer.

        The operation is made of these steps:
        1. A global repartition operation that uses the extractor
        2. A local localgroupby operation that uses the batch reducer

        Once finished, the operation locally outputs a record for each key
        and its respective accumulator value.

        Args:
            extractor (Extractor[Key]): Key extractor function.

            reducer (BatchReducer[T]): Batch reducer function.

        Returns:
            A gear flow outputting a record for each key
        and its respective accumulator value.
        """
        self.operations.append(
            gearop.BatchGroupBy(extractor=extractor, reducer=reducer)
        )
        return self

    def sort(self, reverse: bool = True):
        """Sorting the records

        Note: Using this operation may cause an increase in memory usage
        during runtime due to the list being copied during the sorting
        operation.

        It accepts a single Boolean argument that determines the order.

        The operation is made of the following steps:
        1. A global aggregate operation collects and combines all records
        2. A local sort is performed on the list
        3. The list is flatmapped to records

        Args:
            reverse (bool, optional): Sort in descending order.
            Defaults to True.

        Returns:
            A gear flow outputting the input recods in sorted order.
        """
        self.operations.append(gearop.Sort(reverse=reverse))
        return self

    def distinct(self):
        """Keep only the distinct values in the data.

        It requires no arguments.

        The operation is made of the following steps:
        1. A aggregate operation locally reduces the records to sets that
        are then collected and unionized globally
        2. A local flatmap operation turns the set into records

        Returns:
            A gear flow outputting the unique records, discaring duplicates
        """
        self.operations.append(gearop.Distinct())
        return self

    def count(self):
        """Count the number of records in the execution

        It requires no arguments.

        The operation is made of an aggregate operation that uses local
        counting and global summing accumulators.

        Returns:
            A gear flow outputting the number of input recods (int)
        """
        self.operations.append(gearop.Count())
        return self

    def countby(self, extractor: Extractor[Key] = lambda x: x):
        """Counts the records grouped by key.

        It requires a single extractor function callback.

        The operation is made of an aggregateby operation that uses
        local counting and global summing accumulators.

        Args:
            extractor (Extractor[Key], optional): A function that
            get as input the record and return the key by which to
            performthe counting.
            Defaults to identity function  'lambda x: x'.

        Returns:
            A Gear flow outputting the number of items
            per extracted value.
        """
        self.operations.append(gearop.CountBy(extractor=extractor))
        return self

    def avg(self, extractor: Extractor[Key] = lambda x: float(x)):
        """Calculating arithmetic average of the records

        It accepts an optional value extractor function callback.

        The operation is made of the following steps:
        1. A aggregate operation locally reduces the records to
        tuples of sum and count that are globally combined.
        2. A local map operation calculates the average
        from the global tuple

        Args:
            extractor (Extractor[Key], optional): A function that
            gets the record and return the value by which to
            calculate the average.
            Defaults to identity function  'lambda x: float(x)'.

        Returns:
            [float]: A gear flow outputting the average
            of the extracted values
        """
        self.operations.append(gearop.Avg(extractor=extractor))
        return self


GB = GearsBuilder



import time
import configargparse
import re
from pathlib import Path
import logging
import redgrease.formatting
import redgrease.loader

args = configargparse.ArgParser(
    description="Scans one or more directories for Redis Gears scripts, and "
    "executes them in a Redis Gears instance or cluster. "
    "Can optionally run continiously, montoring and re-loading scripts "
    "whenever changes are detected.",
    default_config_files=["./*.conf", "/etc/redgrease/conf.d/*.conf"],
)

args.add_argument(
    "directories",
    metavar="dir_path",
    nargs="+",
    type=Path,
    help="One or more directories containing Redis Gears scripts to watch",
)

args.add_argument(
    "-c",
    "--config",
    env_var="CONFIG_FILE",
    metavar="PATH",
    is_config_file=True,
    help="Config file path",
)
args.add_argument(
    "--index-prefix",
    env_var="INDEX_PREFIX",
    metavar="PREFIX",
    required=False,
    type=str,
    default=redgrease.loader.default_index_prefix,
    help="Redis key prefix added to the index of monitored/executed script " "files.",
)
args.add_argument(
    "-r",
    "--recursive",
    env_var="RECURSIVE",
    action="store_true",
    help="Recursively watch subdirectories.",
)
args.add_argument(
    "--script-pattern",
    env_var="SCRIPT_PATTERN",
    metavar="PATTERN",
    required=False,
    type=str,
    default=redgrease.loader.default_script_pattern,
    help="File name pattern (glob-style) that must be matched for scripts to "
    "be loaded.",
)
args.add_argument(
    "--requirements-pattern",
    env_var="REQUIREMENTS_PATTERN",
    metavar="PATTERN",
    required=False,
    type=str,
    default=redgrease.loader.default_requirements_pattern,
    help="File name pattern (glob-style) that must be matched for requirement "
    "files to be loaded.",
)
args.add_argument(
    "--unblocking-pattern",
    env_var="UNBLOCKING_PATTERN",
    metavar="PATTERN",
    required=False,
    type=re.compile,
    default=redgrease.loader.default_unblocking_pattern,
    help="Scripts with file paths that match this regular expression, "
    "will be executed with the 'UNBLOCKING' modifier, i.e. async execution. "
    "Note that the pattern is a 'search' pattern and not anchored to the"
    "start of the path string.",
)
args.add_argument(
    "-i",
    "--ignore",
    env_var="IGNORE",
    metavar="PATTERN",
    action="append",
    required=False,
    help="Ignore files matching this pattern.",
)
args.add_argument(
    "-w",
    "--watch",
    env_var="WATCH",
    metavar="SECONDS",
    nargs="?",
    type=float,
    default=0.0,
    const=5.0,
    help="If set, the directories will be continously montiored for updates/"
    "modifications to scripts and requirement files, and automatically loaded/"
    "rerun. The flag takes an optional value specifying the  duration, "
    "in seconds, to wait for further updates/modifications to files, before "
    "executing. "
    "This 'hysteresis' period is to prevent malformed scripts to be "
    "unnecessarily loaded during coding. "
    "If no value is supplied, the duration is defaulting to 5 seconds.",
)
args.add_argument(
    "-s",
    "--server",
    env_var="SERVER",
    type=str,
    const="redis",
    default="localhost",
    nargs="?",
    help="Redis Gears host server IP or hostname.",
)
args.add_argument(
    "-p",
    "--port",
    env_var="PORT",
    type=int,
    default=6379,
    help="Redis Gears host port number",
)
args.add_argument(
    "-l",
    "--log-config",
    env_var="LOG_CONFIG",
    type=str,
    help="",
)


def blah(
    apanapansson: int = 2, bepanbepansdotter: str = "wut teh hellz", some_value=None
):
    ...


config = args.parse_args()

redgrease.formatting.initialize_logger(config=config.log_config)

log = logging.getLogger(__name__)


def main():
    log.debug(config)
    loader = redgrease.loader.GearsLoader(
        script_pattern=config.script_pattern,
        requirements_pattern=config.requirements_pattern,
        unblocking_pattern=config.unblocking_pattern,
        ignore_patterns=config.ignore,
        index_prefix=config.index_prefix,
        server=config.server,
        port=config.port,
        observe=config.watch,
    )
    for directory in config.directories:
        loader.add_directory(directory=directory, recursive=config.recursive)

    if loader.observer:
        log.info("Starting directory observer!")
        loader.start()
        running = True

        try:
            while running:
                time.sleep(1)
        except KeyboardInterrupt:
            running = False
            log.warning("Interrupted by user. Ending!")
        finally:
            loader.stop()


if __name__ == "__main__":
    main()



from redis import Redis
from typing import Optional, Union, List, Iterable, Mapping
import attr
from enum import Enum
import redgrease
import logging

log = logging.getLogger(__name__)


def to_bool(input):
    if isinstance(input, bytes):
        input = input.decode()
    if isinstance(input, str):
        return input.lower() in ["true", "yes", "ok", "1", "y", "ja", "oui"]
    return bool(input)


def to_bytes(input):
    if isinstance(input, bytes):
        return input

    if isinstance(input, str):
        return input.encode()

    if isinstance(input, ExecID):
        return input.id

    if hasattr(input, "__bytes__"):
        input.__bytes__()

    raise ValueError(
        f"Value {input} :: {type(input)} is not a valid '{attr}' as bytes."
    )


def to_str(input):
    return input.decode() if isinstance(input, bytes) else str(input)


def to_list(mapping: dict):
    return list([item for kwpair in mapping.items() for item in kwpair])


def to_dict(
    items: Iterable,
    keyname: str = None,
    keytype=lambda x: x,
    valuename: str = None,
    valuetype=lambda x: x,
):
    kwargs = {}
    iterator = iter(items)
    key_is_set = False
    value_is_set = False
    for item in iterator:
        if not key_is_set:
            if keyname is None:
                key = keytype(item)
                key_is_set = True
            else:
                if to_str(item) == to_str(keyname):
                    item = next(iterator)
                    key = keytype(item)
                    key_is_set = True
        elif not value_is_set:
            if valuename is None:
                value = valuetype(item)
                value_is_set = True
            else:
                if to_str(item) == to_str(valuename):
                    item = next(iterator)
                    value = valuetype(item)
                    value_is_set = True

        if key_is_set and value_is_set:
            kwargs[key] = value
            key_is_set, value_is_set = False, False

    return kwargs


def to_kwargs(items):
    return to_dict(items, keytype=to_str)


def list_parser(item_parser):
    def parser(input_list):
        return list(map(item_parser, input_list))

    return parser


def ok(command_result):
    return True if command_result.startswith(b"OK") else False


@attr.s(frozen=True, auto_attribs=True)
class ExecID:
    shard_id: str = "0000000000000000000000000000000000000000"
    sequence: int = 0

    def __str__(self):
        return f"{self.shard_id}-{self.sequence}"

    def __bytes__(self):
        return str(self).encode()

    def __repr__(self):
        class_name = self.__class__.__name__
        return (
            f"{class_name}(" f"shard_id={self.shard_id}," f"sequence={self.sequence})"
        )

    @staticmethod
    def parse(value):
        if isinstance(value, bytes):
            value = value.decode()

        values = value.split("-")

        shard_id = values[0]
        sequence = int(values[1])

        return ExecID(shard_id=shard_id, sequence=sequence)


class REnum(Enum):
    def __str__(self):
        return str(self.value)

    def __bytes__(self):
        return bytes(self.value)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.value})"


class ExecutionStatus(REnum):
    created = b"created"
    running = b"running"
    done = b"done"
    aborted = b"aborted"
    pending_cluster = b"pending_cluster"
    pending_run = b"pending_run"
    pending_receive = b"pending_receive"
    pending_termination = b"pending_termination"


class ExecLocality(REnum):
    Shard = "Shard"
    Cluster = "Cluster"


class RedisObject:
    @classmethod
    def from_redis(cls, params):
        return cls(**to_kwargs(params))


# @dataclass
@attr.s(auto_attribs=True, frozen=True)
class ExecutionInfo(RedisObject):
    executionId: ExecID = attr.ib(converter=ExecID.parse)
    status: ExecutionStatus = attr.ib(converter=ExecutionStatus)
    registered: bool


# @dataclass
@attr.s(auto_attribs=True, frozen=True)
class RegData(RedisObject):
    mode: redgrease.TriggerMode = attr.ib()
    numTriggered: int
    numSuccess: int
    numFailures: int
    numAborted: int
    lastError: str
    args: dict = attr.ib(converter=to_kwargs)


# @dataclass
@attr.s(auto_attribs=True, frozen=True)
class Registration(RedisObject):
    id: ExecID = attr.ib(converter=ExecID.parse)
    reader: redgrease.Reader
    desc: str
    RegistrationData: RegData = attr.ib(converter=RegData.from_redis)
    PD: dict


@attr.s(auto_attribs=True, frozen=True)
class ExecutionStep(RedisObject):
    type: str = attr.ib(converter=to_str)
    duration: int
    name: str = attr.ib(converter=to_str)
    arg: str = attr.ib(converter=to_str)


@attr.s(auto_attribs=True, frozen=True)
class ExecutionPlan(RedisObject):
    status: ExecutionStatus  # = attr.ib(converter=ExecutionStatus)
    shards_received: int
    shards_completed: int
    results: int
    errors: int
    total_duration: int
    read_duration: int
    steps: List[ExecutionStep] = attr.ib(
        converter=list_parser(ExecutionStep.from_redis)
    )

    @classmethod
    def parse(res):
        executions = map(
            lambda x: to_dict(
                x,
                keyname="shard_id",
                valuename="execution_plan",
                valuetype=ExecutionPlan.from_redis,
            ),
            res,
        )
        return {k: v for d in executions for k, v in d.items()}


@attr.s(auto_attribs=True, frozen=True)
class ShardInfo(RedisObject):
    id: str = attr.ib(converter=to_str)
    ip: str = attr.ib(converter=to_str)
    port: int
    unixSocket: str = attr.ib(converter=to_str)
    runid: str = attr.ib(converter=to_str)
    minHslot: int
    maxHslot: int


@attr.s(auto_attribs=True, frozen=True)
class ClusterInfo(RedisObject):
    my_id: str
    shards: List[ShardInfo] = attr.ib(converter=list_parser(ShardInfo.from_redis))

    @classmethod
    def parse(res):
        if res is None or res == b"no cluster mode":
            return None

        cluster_info = ClusterInfo(
            my_id=to_str(res[1]),
            shards=res[2],  # list(map(ShardInfo.from_redis, res[2]))
        )

        return cluster_info


@attr.s(auto_attribs=True, frozen=True)
class PyStats(RedisObject):
    TotalAllocated: int
    PeakAllocated: int
    CurrAllocated: int


@attr.s(auto_attribs=True, frozen=True)
class PyRequirementInfo(RedisObject):
    GearReqVersion: int
    Name: str = attr.ib(converter=to_str)
    IsDownloaded: bool = attr.ib(converter=to_bool)
    IsInstalled: bool = attr.ib(converter=to_bool)
    CompiledOs: str = attr.ib(converter=to_str)
    Wheels: List[str] = attr.ib(
        converter=lambda wheels: to_str(wheels)
        if isinstance(wheels, bytes)
        else [to_str(wheel) for wheel in wheels]
    )


class RedisGears(Redis):
    class ConfigKey(REnum):
        MaxExecutions = b"MaxExecutions"
        MaxExecutionsPerRegistration = b"MaxExecutionsPerRegistration"
        ProfileExecutions = b"ProfileExecutions"
        PythonAttemptTraceback = b"PythonAttemptTraceback"
        DownloadDeps = b"DownloadDeps"
        DependenciesUrl = b"DependenciesUrl"
        DependenciesSha256 = b"DependenciesSha256"
        PythonInstallationDir = b"PythonInstallationDir"
        CreateVenv = b"CreateVenv"
        ExecutionThreads = b"ExecutionThreads"
        ExecutionMaxIdleTime = b"ExecutionMaxIdleTime"
        PythonInstallReqMaxIdleTime = b"PythonInstallReqMaxIdleTime"
        SendMsgRetries = b"SendMsgRetries"

    # TODO: Use callbacks from 'redis.utils'
    # TODO: and/or 'redis.client' where applicable
    RESPONSE_CALLBACKS = {
        **Redis.RESPONSE_CALLBACKS,
        **{
            "RG.ABORTEXECUTION": ok,
            # 'RG.CONFIGGET': None,
            "RG.CONFIGSET": lambda res: all(map(ok, res)),
            "RG.DROPEXECUTION": ok,
            "RG.DUMPEXECUTIONS": lambda res: list(map(ExecutionInfo.from_redis, res)),
            "RG.DUMPREGISTRATIONS": lambda res: list(map(Registration.from_redis, res)),
            "RG.GETEXECUTION": ExecutionPlan.parse,
            # 'RG.GETRESULTS': None,
            # 'RG.GETRESULTSBLOCKING': None,
            "RG.INFOCLUSTER": ClusterInfo.parse,
            "RG.PYEXECUTE": lambda res: "OK" if res == b"OK" else res,
            "RG.PYSTATS": PyStats.from_redis,
            "RG.PYDUMPREQS": lambda res: list(map(PyRequirementInfo.from_redis, res)),
            "RG.REFRESHCLUSTER": to_bool,
            # 'RG.TRIGGER': None,
            "RG.UNREGISTER": to_bool,
        },
    }

    def abortexecution(self, id: Union[ExecID, bytes, str]) -> bool:
        """Abort the execution of a function mid-flight

        Args:
            id (Union[ID, bytes, str]): The execution id to abort

        Returns:
            [bool]: True or an error if the execution does not exist or had
            already finished.
        """
        return self.execute_command("RG.ABORTEXECUTION", id)

    # TODO: Problematic name, as it is very similar to Redis' config_get
    def configget(self, *config_name: Union[ExecID, bytes, str]) -> List:
        """Get the value of one or more built-in configuration or
        a user-defined options.

        Args:
            config_name (Union[ID, bytes, str]): One or more names/key
            of configurations to get

        Returns:
            List of config values
        """
        return self.execute_command("RG.CONFIGGET", *config_name)

    # TODO: Problematic name, as it is very similar to Redis' config_set
    def configset(self, **config_setting) -> bool:
        """Set a value of one ore more built-in configuration or
        a user-defined options.

        Args:
            config_setting: Key-value-pairs of config settings

        Returns:
            bool: True if all was successful, false oterwise
        """
        return self.execute_command("RG.CONFIGSET", *to_list(config_setting))

    def dropexecution(self, id: Union[ExecID, bytes, str]) -> bool:
        """
        Remove the execution of a function from the executions list.

        Args:
            id (Union[ID, bytes, str]): Execution ID to remove

        Returns:
            bool: True if successful, or an error if the execution
            does not exist or is still running.
        """
        return self.execute_command("RG.DROPEXECUTION", id)

    def dumpexecutions(self) -> List[ExecutionInfo]:
        """Get list of function executions.
        The executions list's length is capped by the 'MaxExecutions'
        configuration option.

        Returns:
            List[ExecutionInfo]: A list of ExecutionInfo,
            with an entry per execution.
        """
        return self.execute_command("RG.DUMPEXECUTIONS")

    def dumpregistrations(self) -> List[Registration]:
        """Get list of function registrations.

        Returns:
            List[Registration]: A list of Registration,
            with one entry per registered function.
        """
        return self.execute_command("RG.DUMPREGISTRATIONS")

    def getexecution(
        self,
        id: Union[ExecutionInfo, ExecID, str, bytes],
        locality: Optional[ExecLocality] = None,
    ) -> Mapping[bytes, ExecutionPlan]:
        """Get the execution plan details for a function in the execution list.

        Args:
            id (Union[ExecutionInfo, ExecID, str, bytes]):
            Execution identifier for the fuction to fetch execution plan for.
            locality (Optional[ExecLocality], optional): Set to 'Shard' to get
            only local execution pland and set to 'Cluster' to collect
            executions from all shards.
            Defaults to 'Shard' in stand-alone mode,
            but "Cluster" in cluster mode.

        Returns:
            Mapping[bytes, ExecutionPlan]: A dict, mapping cluster ID to
            ExecutionPlan
        """
        if isinstance(id, ExecutionInfo):
            id = id.executionId

        locality = [] if locality is None else [to_str(locality).upper()]
        return self.execute_command("RG.GETEXECUTION", to_bytes(id), *locality)

    def getresults(
        self,
        id: Union[ExecutionInfo, ExecID, str, bytes],
    ):
        """Get the results and errors from the execution details of a function
        that's in the execution list.

        Args:
            id (Union[ExecutionInfo, ExecID, str, bytes]):
            Execution identifier for the fuction to fetch the results
            and errors for.

        Returns:
            Tuple: A tuple of results and errors from the gears function.
        """
        if isinstance(id, ExecutionInfo):
            id = id.executionId

        return self.execute_command("RG.GETRESULTS", id)

    def getresultsblocking(self, id: Union[ExecutionInfo, ExecID, str, bytes]):
        """Get the results and errors from the execution details of a function.
        If the execution is not finished, the call is blocked until execution
        ends.

        Args:
            id (Union[ExecutionInfo, ExecID, str, bytes]):
            Execution identifier for the fuction to fetch the results
            and errors for.

        Returns:
            Tuple: A tuple of results and errors from the gears function
        """
        if isinstance(id, ExecutionInfo):
            id = id.executionId

        return self.execute_command("RG.GETRESULTSBLOCKING", id)

    def infocluster(self) -> ClusterInfo:
        """Gets information about the cluster and its shards.

        Returns:
            ClusterInfo: Cluster information or None if not ion cluster mode.
        """
        return self.execute_command("RG.INFOCLUSTER")

    def pyexecute(
        self, function_string: str, unblocking=False, requirements: List[str] = None
    ):
        """Execute Python code

        Args:
            function_string (str): Serialized Gears Python function

            unblocking (bool, optional): Execute function without waiting for
            it to finish, before returnining.
            Defaults to False. I.e. block until the function returns or fails.

            requirements (List[str], optional): List of python package
            dependencies that the function requires in order to execute.
            These packages will be download if not already installed.
            Defaults to None.

        Returns:
            An error is returned if the function can't be parsed, as well as
            any that are generated by non-RedisGears functions used.

            When used in UNBLOCKING mode reply is an execution ID .

            Any results and errors generated by the function are returned as
            an array made of two sub-arrays: one for results and the other for
            errors.

            A simple 'OK' string is returned if the function has no output
            (i.e. it doesn't consist of any functions with the run action).
        """
        params = []
        if unblocking:
            params.append("UNBLOCKING")

        if requirements is not None:
            params.append("REQUIREMENTS")
            params += requirements

        return self.execute_command("RG.PYEXECUTE", function_string, *params)

    def pystats(self) -> PyStats:
        """Get memory usage statisticy from the Python interpreter

        Returns:
            PyStats: Python interpretere memory statistics, including total,
            peak and current amount of allocated memory, in bytes.
        """
        return self.execute_command("RG.PYSTATS")

    def pydumpreqs(self) -> List[PyRequirementInfo]:
        """Gets all the python requirements available (with information about
        each requirement).

        Returns:
            List[PyRequirementInfo]: List of Python requirement information
            objects.
        """
        return self.execute_command("RG.PYDUMPREQS")

    def refreshcluster(self) -> bool:
        """Refreshes the local node's view of the cluster topology.

        Returns:
            bool: True if successful, raises an error othewise
        """
        return self.execute_command("RG.REFRESHCLUSTER")

    def trigger(self, trigger_name: str, *args) -> List:
        """Trigger the execution of a registered 'CommandReader' function.

        Args:
            trigger_name (str): The registered 'trigger' name of the function

            *args (Any): Any additional arguments to the trigger

        Returns:
            List: A list of the functions output records.
        """
        return self.execute_command("RG.TRIGGER", to_str(trigger_name), *args)

    def unregister(self, id: Union[ExecutionInfo, ExecID, str, bytes]) -> bool:
        """Removes the registration of a function

        Args:
            id (Union[ExecutionInfo, ExecID, str, bytes]):
            Execution identifier for the function to unregister.

        Returns:
            bool: True or an error.
            An error is returned if the registration ID doesn't exist or if
            the function's reader doesn't support the unregister operation.
        """
        if isinstance(id, ExecutionInfo):
            id = id.executionId

        return self.execute_command("RG.UNREGISTER", id)



from fnmatch import fnmatch
from datetime import datetime
from os.path import isfile
from typing import Union
from pathlib import Path
import re
import logging

from redis.exceptions import ResponseError

from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler

from redgrease import client, requirements, hysteresis, formatting


log = logging.getLogger(__name__)

default_index_prefix = "/redgrease/scripts"
default_script_pattern = "*.py"
default_requirements_pattern = "*requirements*.txt"
default_unblocking_pattern = "unblock"
default_ignore_patterns = []


def fail(exception, *messages):
    """Convenience function for raising exceptions

    Args:
        exception ([Exception]): Exception to raise

        *messages (str): Any number of messages.
        Will be separated by full stop/period (.).

    Raises:
        exception: Exception
    """
    error_message = ". ".join(messages)
    log.error(error_message)

    raise exception(error_message)


class GearsLoader:
    def __init__(
        self,
        script_pattern: str = None,
        requirements_pattern: str = None,
        unblocking_pattern: Union[str, re.Pattern] = None,
        ignore_patterns: str = None,
        index_prefix: str = None,
        server: str = "localhost",
        port: int = 6379,
        observe: float = None,
        **redis_kwargs,
    ):
        self.directories = []

        self.script_pattern = (
            script_pattern if script_pattern else default_script_pattern
        )

        self.requirements_pattern = (
            requirements_pattern
            if requirements_pattern
            else default_requirements_pattern
        )

        if unblocking_pattern is None:
            unblocking_pattern = default_unblocking_pattern
        if isinstance(unblocking_pattern, str):
            unblocking_pattern = re.compile(unblocking_pattern)
        if not isinstance(unblocking_pattern, re.Pattern):
            fail(ValueError, f"Invalid unblocking pattern: {unblocking_pattern}")
        self.unblocking_pattern = unblocking_pattern

        self.ignore_patterns = (
            ignore_patterns if ignore_patterns else default_ignore_patterns
        )

        self.index_prefix = index_prefix if index_prefix else default_index_prefix

        self.redis = client.RedisGears(host=server, port=port, **redis_kwargs)

        self.observer = Observer() if observe else None

        self.file_index = hysteresis.HysteresisHandlerIndex(hysteresis_duration=observe)

        # Create the file change event listener
        self.event_handler = (
            PatternMatchingEventHandler(
                patterns=[self.script_pattern, self.requirements_pattern],
                ignore_patterns=self.ignore_patterns,
                ignore_directories=False,
                case_sensitive=False,
            )
            if self.observer
            else None
        )

        if self.event_handler:
            self.event_handler.on_deleted = self.on_deleted
            self.event_handler.on_modified = self.on_modified
            self.event_handler.on_moved = self.on_moved

    def __del__(self):
        try:
            self.redis.close()
        except Exception as ex:
            log.warn(f"Error while closing redis: {ex}")

        try:
            self.stop()
        except Exception as ex:
            log.warn(f"Error stopping observer: {ex}")

    def add_directory(self, directory: Union[Path, str], recursive: bool = False):
        if not isinstance(directory, Path):
            directory = Path(str(directory))

        log.info(f"Adding event handlers for {directory}")

        # select either recursive or non-recursive glob function for the dir
        globber = directory.rglob if recursive else directory.glob

        # find and run/register all requirements files in the watch directories
        for requirements_file_path in globber(self.requirements_pattern):
            self.update_dependencies(requirements_file_path)

        # find and run/register all script files in the watch directory
        for script_file_path in globber(self.script_pattern):
            self.register_script(script_file_path)

        if self.observer is not None:
            self.observer.schedule(self.event_handler, directory, recursive)

        self.directories.append(directory)

    # Execute / Register script in redis
    # TODO: Should be integrated into redgrease.RedisGears class
    # TODO: Handle multiple registrations per script.
    def register_script(self, script_path):
        """Execute / Register a gear script in redis using 'RG.PYEXECUTE'

        Args:
            script_path (str): Path to the script.

            Note: paths that maches the 'unblocking-pattern', will be executed
            with the 'UNBLOCKING' modifier.
        """
        log.debug(f"Registering script '{script_path}'")
        general_failure_msg = f"Unable to register script file '{script_path}'"

        if not isfile(script_path):
            fail(FileNotFoundError, general_failure_msg, "File not found")

        with open(script_path) as script_file:
            script_content = script_file.read()

        if not script_content:
            fail(IOError, general_failure_msg, "File is empty")

        # Unregister script if already present
        self.unregister_script(script_path)

        unblocking = self.unblocking_pattern.search(str(script_path)) is not None

        try:
            log.debug(
                f"Running/registering Gear script '{script_path}' " "on Redis server"
            )

            # This is a quite unsafe way of checking for registrations
            # Probabls Ok for dev situations in non-shared environments
            pre_reg = self.redis.dumpregistrations()
            exec_res = self.redis.pyexecute(script_content, unblocking=unblocking)
            ret = f"with return code '{exec_res}'."
            post_reg = self.redis.dumpregistrations()
            pre_reg = set([reg.id for reg in pre_reg])
            log.debug(f"Pre regs: {list(pre_reg)}")
            post_reg = set([reg.id for reg in post_reg])
            log.debug(f"Post regs: {list(post_reg)}")
            diff_reg: set = post_reg - pre_reg
            if len(diff_reg) > 0:
                reg_id = diff_reg.pop()
                self.redis.hset(
                    f"{self.index_prefix}{script_path}",
                    mapping={
                        "registration_id": str(reg_id),
                        "last_updated": datetime.utcnow().strftime(
                            formatting.iso8601_datefmt
                        ),
                    },
                )
                log.debug(f"Script '{script_path}' registered as '{reg_id}' {ret}")
            else:
                log.debug(f"Script '{script_path}' executed {ret}")

            if len(diff_reg) > 0:
                log.warning(
                    "Multiple registrations occured? "
                    "Index migth be corrupt. "
                    "Is this a shared environment?"
                )

        except Exception as ex:
            log.error(f"Something went wrong: {ex}")

    # Actions
    # TODO: Should be integrated into redgrease.RedisGears class
    # TODO: Handle multiple registrations per script.
    def unregister_script(self, script_path):
        """Check if a given script is registered, and if so,
        unregister it using 'RG.UNREGISTER'

        Args:
            script_path (str): Script path
        """
        log.debug(f"Unregistering script: '{script_path}'")
        reg_key = f"{self.index_prefix}{script_path}"
        reg_id = self.redis.hget(reg_key, "registration_id")
        if reg_id is not None:
            log.debug(
                f"Removing registration for script '{script_path}' "
                f"with id '{reg_id}'"
            )
            try:
                self.redis.unregister(reg_id)
            except ResponseError as err:
                log.warn(
                    "Unregistration failed. "
                    "Index migth be corrupt. "
                    "Is this a shared environment?"
                )
                log.error({err})
        self.redis.delete(reg_key)

    def update_dependencies(self, requirements_file_path):
        """Update (add only) package dependencies on the Redis instance

        Args:
            requirements_file_path (str): File path of 'requirements.txt' file.
        """
        log.debug(f"Updating dependecies as per '{requirements_file_path}'")
        try:
            requirements_list = list(requirements.read(requirements_file_path))
            log.debug(f"Requirements to load: {', '.join(requirements_list)}")
            self.redis.pyexecute("GB().run()", requirements=requirements_list)
        except Exception as ex:
            log.error(f"Something went wrong: {ex}")

    # File Event Handling
    def on_deleted(self, event):
        """Watchdog event handler for events signalling that a
        script or requirement file has been deleted.
        Script files are unregistered (if present) and re-run,
        after applying some hystersis.
        Removed requirement files does not remove any installed packages.

        Args:
            event (watchdog.FileDeletedEvent): Event data for deleted file
        """
        file = event.src_path
        if fnmatch(file, self.script_pattern):
            log.debug(
                f"Gears script '{file}' deleted. "
                "Scheduling unregistration of script."
            )
            # Apply hysteresis in case additional events shortly follow,
            # before we actually unregister
            self.file_index.signal(file, self.unregister_script, file)
        elif fnmatch(file, self.requirements_pattern):
            log.info(
                f"Gears requirements file '{file}' deleted. "
                "Requirement removal not Implemented. "
                "Ignoring."
            )
        else:
            log.warn(f"Unknown file type '{file}' deleted. " "Ignoring.")

    def on_modified(self, event):
        """Watchdog event handler for events signalling that a
        script or requirement file has been modified.

        Args:
            event ([type]): [description]
        """
        file = event.src_path
        if fnmatch(file, self.script_pattern):
            log.debug(f"Gears script '{file}' modified. Regestering script.")
            # Apply hysteresis in case additional events shortly follow,
            # before we actually (re-)run/register the script.
            self.file_index.signal(file, self.register_script, file)
        elif fnmatch(event.src_path, self.requirements_pattern):
            log.debug(
                f"Gears requirements file '{file}' modified. " "Updating dependencies."
            )
            # Apply hysteresis in case additional events shortly follow,
            # before we actually update requirements.
            self.file_index.signal(file, self.update_dependencies, file)
        else:
            log.warn(f"Unknown file type '{event.src_path}' modified. Ignoring.")

    def on_moved(self, event):
        """Watchdog event handler for events signalling that a
        script or requirement file has been moved.
        Script files are unregistered (if present) and re-run,
        after applying some hystersis.
        Removed requirement files does not remove any installed packages.

        Args:
            event ([type]): [description]
        """
        old_file = event.src_path
        new_file = event.dest_path
        # Handle, i.e. unregister, the old / source file.
        if fnmatch(old_file, self.script_pattern):
            log.debug(
                f"Gears script '{old_file}' moved to '{new_file}'. "
                "Unregistering old script."
            )
            # Apply hysteresis in case additional events shortly follow,
            # before we actually unregister the script.
            self.file_index.signal(old_file, self.unregister_script, old_file)
        elif fnmatch(old_file, self.requirements_pattern):
            log.debug(
                f"Gears requirements file '{old_file}' moved to '{new_file}'. "
                "Requirement removal not Implemented. "
                "Ignoring."
            )

        # TODO: Check that the new file is in the watch directory
        # Handle, i.e. run/register, the new / destination file
        if fnmatch(new_file, self.script_pattern):
            log.debug(
                f"File '{old_file}' moved to Gears script '{new_file}'. "
                "Registering new script."
            )
            # Apply hysteresis in case additional events shortly follow,
            # before we actually (re-)run/register the script
            self.file_index.signal(new_file, self.register_script, new_file)
        elif fnmatch(new_file, self.requirements_pattern):
            log.debug(
                f"File '{old_file}' moved to requirements file '{new_file}'. "
                "Updating dependencies"
            )
            # Apply hysteresis in case additional events shortly follow,
            # before we actually update requirements.
            self.file_index.signal(new_file, self.update_dependencies, new_file)
        else:
            log.warn(
                f"File '{old_file}' moved to unknonwn type '{new_file}'. " "Ignoring."
            )

    def start(self):
        log.info("Starting RedGrease directory watcher ")
        if self.observer:
            self.observer.start()

    def stop(self):
        if self.observer:
            self.observer.stop()



from dataclasses import dataclass, field
from redgrease.sugar import TriggerMode
from redgrease.typing import T, Key
import redgrease.typing as optype


# Redis Gears Operations
# Should maybe be in a separate sub-package 'redgrease.operations'
@dataclass
class Operation:
    pass


@dataclass
class Run(Operation):
    arg: str = None
    convertToStr: bool = True
    collect: bool = True
    kargs: dict = field(default_factory=dict)


@dataclass
class Register(Operation):
    prefix: str = "*"
    convertToStr: bool = True
    collect: bool = True
    mode: str = TriggerMode.Async
    onRegistered: optype.Callback = None
    trigger: str = None
    kargs: dict = field(default_factory=dict)


@dataclass
class Map(Operation):
    op: optype.Mapper


@dataclass
class FlatMap(Operation):
    op: optype.Expander


@dataclass
class ForEach(Operation):
    op: optype.Processor


@dataclass
class Filter(Operation):
    op: optype.Filterer


@dataclass
class Accumulate(Operation):
    op: optype.Accumulator


@dataclass
class LocalGroupBy(Operation):
    extractor: optype.Extractor[Key]
    reducer: optype.Reducer


@dataclass
class Limit(Operation):
    length: int
    start: int


@dataclass
class Collect(Operation):
    pass


@dataclass
class Repartition(Operation):
    extractor: optype.Extractor[Key]


@dataclass
class Aggregate(Operation):
    zero: T
    seqOp: optype.Accumulator[T]
    combOp: optype.Accumulator[T]


@dataclass
class AggregateBy(Operation):
    extractor: optype.Extractor[Key]
    zero: T
    seqOp: optype.Reducer[T]
    combOp: optype.Reducer[T]


@dataclass
class GroupBy(Operation):
    extractor: optype.Extractor[Key]
    reducer: optype.Reducer[T]


@dataclass
class BatchGroupBy(Operation):
    extractor: optype.Extractor[Key]
    reducer: optype.BatchReducer[T]


@dataclass
class Sort(Operation):
    reverse: bool = True


@dataclass
class Distinct(Operation):
    pass


@dataclass
class Count(Operation):
    pass


@dataclass
class CountBy(Operation):
    extractor: optype.Extractor[Key]


@dataclass
class Avg(Operation):
    extractor: optype.Extractor[Key]



import logging

# Various Helper enums, instead of "magic" strings.


class Reader:
    """Redis Gears Reader Types"""

    KeysReader = "KeysReader"
    KeysOnlyReader = "KeysOnlyReader"
    StreamReader = "StreamReader"
    PythonReader = "PythonReader"
    ShardsIDReader = "ShardsIDReader"
    CommandReader = "CommandReader"


class TriggerMode:
    """Redis Geears Trigger modes for registered actions"""

    Async = "async"
    AsyncLocal = "async_local"
    Sync = "sync"


class LogLevel:
    """Redis Gears log levels"""

    Debug = "debug"
    Verbose = "verbose"
    Notice = "notice"
    Warning = "warninig"

    @staticmethod
    def to_logging_level(rg_log_level):
        if rg_log_level == LogLevel.Debug:
            return logging.DEBUG
        elif rg_log_level == LogLevel.Verbose:
            return logging.INFO
        elif rg_log_level == LogLevel.Notice:
            return logging.WARNING
        elif rg_log_level == LogLevel.Warning:
            return logging.ERROR  # Maybe a little unintuitive ;)
        else:
            return logging.INFO



from redgrease import GearsBuilder, TriggerMode
from redgrease.typing import Callback


def trigger(
    trigger: str,
    prefix: str = "*",
    convertToStr: bool = True,
    collect: bool = True,
    mode: str = TriggerMode.Async,
    onRegistered: Callback = None,
    **kargs
):
    def gear(func):
        GearsBuilder("CommandReader").map(lambda params: func(params[1:])).register(
            prefix=prefix,
            convertToStr=convertToStr,
            collect=collect,
            mode=mode,
            onRegistered=onRegistered,
            trigger=trigger,
            **kargs
        )

    return gear



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import setuptools
import pathlib
import sys

here = pathlib.Path(__file__).parent.resolve()


def text(*names, encoding="utf8"):
    return here.joinpath(*names).read_text(encoding=encoding)


print(f"setup file : {__file__}")

# try:
#     version = open(".version").readline()
# except FileNotFoundError:
#     print("Unable to determine version.")
#     print("Set the version on the first line of a '.version' file")
#     sys.exit(404)


runtime_extras = ["redis"]
client_extras = runtime_extras + [
    "attrs",
]
cli_extras = client_extras + [
    "watchdog",
    "ConfigArgParse",
    "pyyaml",
]

all_extras = cli_extras

setuptools.setup(
    name="redgrease",
    #    version=version,
    license="MIT License",
    description="RedisGears helper package",
    long_description=text("README.md"),
    long_description_content_type="text/markdown",
    author="Anders Åström",
    author_email="anders@lyngon.com",
    url="https://github.com/lyngon/redgrease",
    packages=setuptools.find_packages(where="src"),
    package_dir={"": "src"},
    classifiers=[  # https://pypi.org/classifiers/
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Database",
        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries :: Application Frameworks",  # noqa: E501
    ],
    install_requires=[],  # Should remain empty for minimal runtime install.
    extras_require={
        "all": all_extras,
        "client": client_extras,
        "cli": cli_extras,
        "runtime": runtime_extras,
    },
    entry_points={
        "console_scripts": ["redgrease=redgrease.cli:main [cli]"],
    },
    setup_requires=[],
    python_requires=">=3.6",
    keywords="Redis, Gears, development",
)

sys.exit(0)



from redgrease import execute
from redgrease.client import RedisGears
from functools import lru_cache


def runtime_execute_command(self, *args, **options):
    "Execute a command in local Gear runtime and return a parsed response"
    command_name = args[0]
    response = execute(*args)
    if command_name in self.response_callbacks:
        return self.response_callbacks[command_name](response, **options)
    return response


@lru_cache
def get_runtime_client():
    runtime_client = RedisGears(connection_pool=...)
    runtime_client.execute_command = runtime_execute_command
    return runtime_client


execute = get_runtime_client()  # noqa: F811 - Used as exported var


# ## Redis v6 commands
# # Cluster
# Cluster AddSlots
# Cluster BumpEpoch
# Cluster Count-Failure-Reports
# Cluster CountKeysInSlot
# Cluster DelSlots
# Cluster Failover
# Cluster FlushSlots
# Cluster Forget
# Cluster GetKeysInSlot
# Cluster Info
# Cluster KeySlot
# Cluster Meet
# Cluster MyID
# Cluster Nodes
# Cluster Replicate
# Cluster Reset
# Cluster SaveConfig
# Cluster Set-Config-Epoch
# Cluster SetSlot
# Cluster Slaves
# Cluster Replicas
# Cluster Slots
# ReadOnly
# ReadWrite


# # Connection
# Auth
# Client Caching
# Client Id
# Client Info
# Client Kill
# Client List
# Client GetName
# Client GetRedir
# Client Unpause
# Client Pause
# Client Reply
# Client SetName
# Client Tracking
# Client TrackingInfo
# Client Unblock
# Echo
# Reset
# Select

# # Geo
# GeoAdd
# GeoHash
# GeoPos
# GeoDist
# GeoRadius
# GeoRadiusByMember
# GeoSearch
# GeoSearchStore

# # Hashes
# HDel
# HExists
# HGet
# HGetAll
# HIncrBy
# HIncrByFloat
# HKeys
# HLen
# HMGet
# HMSet
# HSet
# HSetNX
# HStrLen
# HVals
# HScan

# # HyperLogLog
# PFAdd
# PFCount
# PFMerge

# # Keys
# Copy
# Del
# Dump
# Exists
# Expire
# ExpireAt
# Keys
# Migrate
# Move
# Object
# Persist
# PExpire
# PExpireAt
# PTTL
# RandomKey
# Rename
# RenameNX
# Restore
# Sort
# Touch
# TTL
# Type
# Unlink
# Wait
# Scan

# # List
# BLPop
# BRPop
# BRPopLPush
# BLMove
# LIndex
# LInsert
# LLen
# LPop
# LPos
# LPush
# LPushX
# LRange
# LRem
# LSet
# LTrim
# RPop
# RPopLPush
# LMove
# RPush
# RPushX

# # Pub/Sub
# PSubscribe
# PubSub
# Publish
# PUnsubscribe
# Subcribe
# Unsubscribe

# # Scripting
# Eval
# EvalSha
# Script Debug
# Script Exists
# Script Flush
# Script Kill
# Script Load

# # Server
# ACL Load
# ACL Save
# ACL List
# ACL Users
# ACL GetUser
# ACL SetUser
# ACL DelUser
# ACL Cat
# ACL GenPass
# ACL WhoAmI
# ACL Log
# ACL Help
# BGRewriteAOF
# BGSave
# Command
# Command Count
# Command GetKeys
# Command Info
# Config Get
# Config Rewrite
# Config Set
# Config ResetStat
# DBSize
# Debug Object
# Debug Segfault
# FlushAll
# FlushDB
# Info
# LolWut
# LastSave
# Memory Doctor
# Memory Help
# Memory Malloc-Stats
# Memory Purge
# Memors Stats
# Memory Usage
# Module List
# Module Load
# Module Unload
# Monitor
# Role
# Save
# Shutdown
# SlaveOf
# ReplicaOf
# SlowLog
# SwapDB
# Sync
# PSync
# Time
# Latency Docto4
# Latency Graph
# Latency History
# Latency Latest
# Latency Reset
# Latency Help

# # Sets
# SAdd
# SCard
# SDiff
# SDiffScore
# SInter
# SInterStore
# SIsMember
# SMIsMember
# SMembers
# SMove
# SPop
# SRandMember
# SRem
# SUnion
# SUnionStore
# SScan

# # Sorted Sets
# BZPopMin
# BZPopMax
# ZAdd
# ZCard
# ZCount
# ZDiff
# ZDiffStore
# ZIncrBy
# ZInter
# ZInterStore
# ZLexCount
# ZPopMax
# ZPopMin
# ZRangeStore
# ZRange
# ZRangeByLex
# ZRevRangeByLex
# ZRangeByScore
# ZRank
# ZRem
# ZRemRangeByLex
# ZRemRangeByRank
# ZRemRangeByScore
# ZRevRange
# ZRevRAngeByScore
# ZRevRank
# ZScore
# ZUnion
# ZMScore
# ZUnionStore
# ZScan

# # Streams
# XInfo
# XAdd
# XTrim
# XDel
# XRange
# XRevRange
# XLen
# XRead
# XGroup
# XReadGroup
# XAck
# XClaim
# XAutoClaim
# XPending


# # Strings
# Append
# Bitcount
# Bitfield
# Bitop
# Bitpos
# Decr
# DecrBy
# Get
# GetBit
# GetRange
# GetSet
# Incr
# IncrBy
# IncrByFloat
# MGet
# MSet
# MSetNx
# PSetEx
# Set
# SetBit
# SetEx
# SetNx
# SetRange
# StrAlgo
# StrLen


# # Transactions
# Discard
# Exec
# Muli
# Unwatch
# Watch
